/******************************************************************************
 $Header: /Repository/apps/firmdown/firmdown.c,v 1.25 2006/11/28 07:53:02 as-ewb Exp $

  Firmware downloader for ASI2400

  2005-07-27
  By default Downloads update image \\asi\\sw2\\dsp\\ax2400\\dsp2400_006.bin"
  Edit line 398 to change the path
  Uncomment lines 564 to enable programming of factory and bootloader images.
  Must be running a2400.c 1.30.2.6 or later

 (C) Copyright AudioScience Inc. 2005
*******************************************************************************/
//#include "stdafx.h"
#include <string.h>
#include <stdlib.h>

#include <hpi.h>
#include <hpios.h>  // for debug
#include <hpidspcd.h>
#include <hpidebug.h>
#include <hpinet.h>
#include <getopt.h>
#include <getopt.h>

typedef unsigned int u_int;

void HPI_MessageEx(HPI_MESSAGEX *phm, HPI_RESPONSEX *phr ,
		unsigned int timeout);

void HpiOs_SetDspCodePath(char * pPath);
HW16 SubsysGetAdapterIndexForMac(HPI_ETHERNET_MAC_ADR mac, HW16 * pIndex);
HW16 SubsysGetAdapterIndexForMacString(const char * smac, HW16 * pIndex);
HPI_ERR AdapterGetMACaddress(HW16 adapterIndex,unsigned char ** pMAC);

typedef void progress_callback(char * state, u_int percent);


enum FIRMWARE_TYPE {UPDATE=1,FACTORY=2,BOOTLOADER=3};

#define BOOTLOADER_OFFSET 0
#define FACTORY_OFFSET 0x08000
#define UPDATE_OFFSET  0x80000

#define MAX_FILES 3
#define MAC_STRING_SIZE 12

#ifndef MAX_PATH
#define MAX_PATH 260
#endif

extern int hpinet_pcap_inum;

HPI_HSUBSYS *phSubSys=0 ;

struct fw_info {
	enum FIRMWARE_TYPE type;
	int offset;
	char path[MAX_PATH];
	char name[12];
};

struct fw_info fw[MAX_FILES]= {
	{UPDATE,    UPDATE_OFFSET    , "dsp2400.bin", "Update"},
	{FACTORY,   FACTORY_OFFSET   , "dsp2400.bin", "Factory"},
	{BOOTLOADER,BOOTLOADER_OFFSET, "boot2400.bin","Bootloader"}
};



/* Option variables */
int verbosity = 0;
int image_sel=0;
int num_files=0;
int shown_help=0;
char szMac[MAC_STRING_SIZE+1]="";
char szFile[MAX_FILES][80]={"","",""};
int adapter_specified=0;

static struct option long_options[] =
    {
		{"interface"    , required_argument, 0, 'i'},
		{"mac"          , required_argument, 0, 'm'},
        {"verbosity"    , required_argument, 0, 'v'},
        {"factory"      , no_argument,0,'f'},
        {"update"       , no_argument,0,'u'},
        {"bootloader"   , no_argument,0,'b'},
        {"help"         , no_argument,0,'h'},
        {0, 0, 0, 0}
    };

const char* short_options="i:m:v:fubh?";

const char* option_help[]=
    {
	"Interface index, 0=prompt user",
	"Mac address of ASI2416 to update",
	"Verbosity",
	"Update factory image",
	"Update update image (danger!)",
	"Update bootloader image (danger!)",
        "Show this text."
    };

void help(char * progname)
{
    int i=0;
    printf("\nUsage - %s [options]"
		" [update image [factory image [bootloader image]]]\n", progname);
    while (long_options[i].name != 0) {
        printf("--%s -%c %s\n",
               long_options[i].name,
               (char)(long_options[i].val),
               option_help[i]);
        i++;
    }
	printf("\n");
}

void parse_options(int argc, char *argv[])
{
    int c;
    /*********** Parse the command line options ***************/
    while (1) {
      //int this_option_optind = optind ? optind : 1;
        int option_index = 0;

        c = getopt_long (argc, argv, short_options,
                         long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 0:
            printf ("option %s", long_options[option_index].name);
            if (optarg)
                printf (" with arg %s", optarg);
            printf ("\n");
            break;
        case 'i':
            hpinet_pcap_inum=atoi(optarg);
            break;
        case 'v':
            verbosity=atoi(optarg);
            break;
		case 'm':
			strncpy(szMac,optarg,MAC_STRING_SIZE);
			adapter_specified=1;
			break;

	case 'u':
		image_sel |= 1;
		break;
	case 'f':
		image_sel |= 2;
		break;
	case 'b':
		image_sel |= 4;
		break;
        case '?':
        case 'h':
			shown_help=1;
            help(argv[0]);
			exit(0);
            break;

        default:
            printf ("?? getopt returned character code 0%o ??\n", c);
        }
    }

    if (optind < argc) {
        // printf ("non-option ARGV-elements: ");
        while ((optind < argc) && (num_files < MAX_FILES))  {
            strncpy( fw[num_files].path, argv[optind++], MAX_PATH-1 );
            printf("%s image is %s\n", fw[num_files].name,fw[num_files].path);
            num_files++;
        }
    }

}

/*=================================================================*/


#ifdef _MSC_VER
void delay(int millisecs)
{
	Sleep(millisecs);
}
#else
#include <time.h>

/*---------------------------------------------------------------------------*/
void delay(long millisecs)
{
	struct timespec poll_interval={0,millisecs*1000000l};
	nanosleep(&poll_interval, 0);
}
#endif

/* max 999 */
#define MSG_DELAY 10
/*=================================================================*/
/**

*/
void HPI_PrintError(HW16 wError, HW16 wSpecificError)
{
	char errbuf[256];

	if (wError==0) {

	} else {
		HPI_GetErrorText( wError, errbuf);
				printf(" Error%s, SE=%d\n",errbuf,wSpecificError);	}
}
/*=================================================================*/
/**

*/
HW32 calc_checksum(HW32 checksum, int count, HW32 * pData)
{	int i;
	for (i=0; i< count; i++)
		checksum ^= pData[i];

	return checksum;
}

/** number of words in code tailer */
#define TAILERSIZE 3
/*=================================================================*/
/**

*/
int DownloadFirmware(HW16 wAdapterIndex,enum FIRMWARE_TYPE ft,char *codepath, progress_callback pc)
{
	short error;
	DSP_CODE DspCode;
	HW16 nBootLoadFamily=0x2400;
	HW32 dwLength;
	u_int checksum;
	u_int total_checksum=0;
	u_int dwStartOffset;
	u_int sequence=0;
	u_int progress=0;
	u_int p_inc;

	delay(MSG_DELAY);

	HpiOs_SetDspCodePath(codepath);


	if ((ft > BOOTLOADER) || (ft < UPDATE))
		return 1;

	dwStartOffset=fw[ft-1].offset;
	printf("Downloading %s image\n",fw[ft-1].name);

	HpiDspCode_Open(nBootLoadFamily,&DspCode,NULL);

	if ((error=HpiDspCode_Open(nBootLoadFamily,&DspCode,NULL))!= 0)
		return( error);

	if (ft==BOOTLOADER) {
		/* Bootloader needs to be raw code image, so discard
		   the three header words.
		   Check length of first block is same as total file length-
		   overhead, to make sure there is only one block
	    */
		HW32 dwLength,dwAddress,dwType;

		HpiDspCode_ReadWord(&DspCode,&dwLength);
		HpiDspCode_ReadWord(&DspCode,&dwAddress);
		HpiDspCode_ReadWord(&DspCode,&dwType);


		if (dwLength != (DspCode.dwBlockLength - 4)) {
			printf("Boot code must be in a single block\n");
			return 1;
		}
		dwLength=DspCode.dwBlockLength-4+TAILERSIZE;
	} else {
		dwLength=DspCode.dwBlockLength+TAILERSIZE;
	}

#define WORDS_PER_BLOCK 256
	p_inc= 10000*WORDS_PER_BLOCK/dwLength;
	if (&pc)
		pc("Erasing",0);

	{
		HPI_MESSAGE hm;
		HPI_RESPONSE hr;
		HPI_InitMessage(&hm, HPI_OBJ_ADAPTER, HPI_ADAPTER_START_FLASH);
		hm.wAdapterIndex=wAdapterIndex;
		hm.u.ax.start_flash.dwKey=0x0a51;
		hm.u.ax.start_flash.dwOffset=dwStartOffset;
		hm.u.ax.start_flash.dwLength=DspCode.dwBlockLength+TAILERSIZE;

		hr.wError=0;
		hr.wSize=sizeof(hr);

		HPI_MessageEx((HPI_MESSAGEX *)&hm,(HPI_RESPONSEX *)&hr,10000); // wait up to 10 sec for flash erase
		error=hr.wError;
		delay(MSG_DELAY);
	}

	while ((DspCode.dwWordCount < DspCode.dwBlockLength) && ! error)
	{
		HPI_MESSAGEX hm;
		HPI_RESPONSE hr;
		HW16 wLength;
		HW32 *pdwCode;
		HW32 dwRemain=DspCode.dwBlockLength-DspCode.dwWordCount;
		int retries;

		wLength = WORDS_PER_BLOCK;
		if (wLength > dwRemain)
			wLength = (unsigned short)dwRemain;

		error=HpiDspCode_ReadBlock(wLength,&DspCode,&pdwCode);
		if (error)
			break;

		checksum = calc_checksum(0, wLength, pdwCode);
		total_checksum ^= checksum;

		HPI_InitMessage( (HPI_MESSAGE *)&hm,  HPI_OBJ_ADAPTER, HPI_ADAPTER_PROGRAM_FLASH);
		hm.msg.wAdapterIndex=wAdapterIndex;
		hm.msg.u.ax.program_flash.wSequence=sequence;
		hm.msg.u.ax.program_flash.dwChecksum=checksum;
		hm.msg.u.ax.program_flash.wLength=wLength;
		hm.msg.wSize=sizeof(HPI_MESSAGE)+wLength*sizeof(HW32);


		memcpy(&hm.data,pdwCode,wLength*sizeof(HW32));

		retries=2;

		while (retries) {
			hr.wError=0;
			hr.wSize=sizeof(hr);
			HPI_MessageEx(&hm,(HPI_RESPONSEX*)&hr,2000);
			delay(MSG_DELAY);

			retries--;
			error=hr.wError;
//			printf("%4d Checksum %08x ",sequence,checksum);

			if (error) {
				HPI_PrintError(error , hr.wSpecificError);
			} else {
//				printf("\n");
				break;
			}
		}

		sequence++;
		progress += p_inc;
		if (&pc)
			pc("Programming",progress/100);
	}


	if (!error) {

		HPI_MESSAGEX hm;
		HPI_RESPONSE hr;
		HW32 * tailer = (HW32 *)&hm.data;

		tailer[0]=total_checksum;
		tailer[1]=DspCode.dwBlockLength;
		tailer[2]=DspCode.dwVersion;

		printf("Total checksum %08x\n",total_checksum);

		HPI_InitMessage( (HPI_MESSAGE *)&hm,  HPI_OBJ_ADAPTER, HPI_ADAPTER_PROGRAM_FLASH);
		hm.msg.wAdapterIndex=wAdapterIndex;
		hm.msg.u.ax.program_flash.wSequence=sequence++;
		hm.msg.u.ax.program_flash.dwChecksum=calc_checksum(0, TAILERSIZE, tailer);
		hm.msg.u.ax.program_flash.wLength=TAILERSIZE;
		hm.msg.wSize=sizeof(HPI_MESSAGE)+TAILERSIZE*sizeof(HW32);

		hr.wSize=sizeof(hr);

		HPI_MessageEx(&hm,(HPI_RESPONSEX*)&hr,2000);
		delay(MSG_DELAY);

		error=hr.wError;
	}
	HpiDspCode_Close(&DspCode);
	if (&pc)
		pc("Done",100);
	return error;
}

/*=================================================================*/
/**
Progress callback
*/

void pc(char * state, u_int percent)
{
	printf("%s %d\n",state, percent);
}
/*=================================================================*/
/**

*/
int main(int argc, char* argv[])
{

	int i=0;
	short error=0;
	HW16 wNumAdapters;
#define MAX_ADAPTERS 20
	HW16 awAdapterList[MAX_ADAPTERS];
	HW16 wAdapterIndex=0;
	HPI_HMIXER hMixer=0;

	printf("FIRMDOWN $Revision: 1.25 $ (c) AudioScience, Inc. 2005\n\n");
	parse_options(argc,argv);

	if (!image_sel && !shown_help)
		printf("To see full option list, use -h option\n\n");

	phSubSys=HPI_SubSysCreate();
	if (!phSubSys) {
		printf("SubSysOpen Failed\n");
		return 1;
	}

	error=HPI_SubSysFindAdapters(phSubSys,&wNumAdapters,awAdapterList,MAX_ADAPTERS);

	printf("Found %d adapters\n",wNumAdapters);
	for (i=0; i<MAX_ADAPTERS; i++) {
		if (awAdapterList[i]) {
			unsigned char * pMAC=NULL;
			int m;

			printf("Adapter %d is ASI%04X MAC=",i,awAdapterList[i]);
			AdapterGetMACaddress(i,&pMAC);
			if (pMAC)
				for (m=0; m<6; m++)
					printf("%02x",pMAC[m]);
			printf("\n");
		}
	}

	// exit if more than one adapter found and which one to program hasn't been
	// specified explicitly
	if (!adapter_specified &&(wNumAdapters > 1))
	{
		printf("\n**** ERROR - more than one ASI2416 found, "
			   "use -a or -m option to select one ***\n");
		exit(0);
	}

	if (szMac[0] != 0) {
			error=SubsysGetAdapterIndexForMacString(szMac, &wAdapterIndex);
			if (error) {
				printf("Couldn't find adapter with mac %s\n",szMac);
				exit(1);
			} else {
				printf("Found adapter %d with MAC address %s\n",wAdapterIndex, szMac);
			}
	}

	if (awAdapterList[wAdapterIndex] == 0) {
		printf("No adapter with index %d\n",wAdapterIndex);
		exit(1);
	}

	//========= adapter open ==========
	error=HPI_AdapterOpen(phSubSys,wAdapterIndex);
	HPI_PrintError(error , 0);

	{ //========= adapter get info ==========
		HW16 wNumOutStreams;
		HW16 wNumInStreams;
		HW16 wVersion;
		HW32 dwSerialNumber;
		HW16 wAdapterType;

		error= HPI_AdapterGetInfo(phSubSys,wAdapterIndex,
			&wNumOutStreams,&wNumInStreams,
			&wVersion,&dwSerialNumber,&wAdapterType);
		HPI_PrintError(error , 0);

		printf("Adapter ID=%4X Index=%d NumOutStreams=%d "
			"NumInStreams=%d S/N=%d\nHw Version %c%d DSP code version %03d\n",
			   wAdapterType, wAdapterIndex,
			   wNumOutStreams, wNumInStreams, dwSerialNumber,
			   ((wVersion>>3)&0xf)+'A',    // Hw version major
			   wVersion&0x7,               // Hw version minor
			   ((wVersion>>13)*100)+((wVersion>>7)&0x3f)  // DSP code version
			  );
	}


	{ //========= query flash version ==========

		HPI_MESSAGE hm;
		HPI_RESPONSE hr;
		int sel;

		HPI_InitMessage( &hm,  HPI_OBJ_ADAPTER, HPI_ADAPTER_QUERY_FLASH);
		hm.wAdapterIndex=wAdapterIndex;

		for (sel=0; sel < 3; sel++) {
//			if (image_sel & (1 << sel))
			{
				hm.u.ax.query_flash.dwOffset=fw[sel].offset;
				hr.wError=0;
				hr.wSize=sizeof(hr);
				HPI_MessageEx((HPI_MESSAGEX*)&hm,(HPI_RESPONSEX *)&hr,1000);
				HPI_PrintError(hr.wError , 0);
				printf("%10s image, version %3d, checksum %08x, length %d\n",
					fw[sel].name,
					hr.u.ax.query_flash.dwVersion,
					hr.u.ax.query_flash.dwChecksum,
					hr.u.ax.query_flash.dwLength);
			}
		}

	}

	//========= program selected images to flash ==========
	for (i=0; i<3; i++) {
		if (image_sel & (1 << i)) {
			error=DownloadFirmware(wAdapterIndex,fw[i].type,fw[i].path,pc);
				HPI_PrintError(error , 0);
		}
	}

	if(hMixer)
		HPI_MixerClose(phSubSys,hMixer);

	if (image_sel) {
		HPI_MESSAGE hm;
		HPI_RESPONSE hr;

		// Send a special close message to the adapter to reset it
		HPI_InitMessage( &hm,  HPI_OBJ_ADAPTER, HPI_ADAPTER_CLOSE);
		hm.wAdapterIndex=wAdapterIndex;
		hm.u.a.wObjectType=0xdead;
		hm.u.a.wAssertId=0x0FF;
		HPI_MessageEx((HPI_MESSAGEX*)&hm,(HPI_RESPONSEX *)&hr,10);
	}

    HPI_DriverClose();

	if (error) {

		exit(1);
	}
	exit(0);
}

// $Log: firmdown.c,v $
// Revision 1.25  2006/11/28 07:53:02  as-ewb
// add MAX_PATH if not defined
//
// Revision 1.24  2006/11/28 04:49:59  as-ewb
// Allow selection of adapter by MAC when more than one adapter present on network.  -a option removed because adapter index potentially can change on each run depending on which adapter answers the broadcast query first.
// Print the MAC addresses of found adapters.
//
// Revision 1.23  2006/11/28 00:38:23  as-ewb
// Add -h option
//
// Revision 1.22  2006/11/27 23:26:38  as-ewb
// Minor change to help text.
// Add -m and -a options for choosing which adapter to program.
// Note selection of what image to program is by a combination of the options -u, -f or -b
//
// Revision 1.21  2006/11/27 22:24:52  as-ewb
// add getopt style options
//
// Revision 1.20  2006/11/16 10:26:24  as-ewb
// Get firmware files from local directory (no path).
// Remove all unused test code.
//
// Revision 1.19  2006/03/14 15:11:06  as-tfe
// Add new parameter to HpiDspCode_Open().
//
// Revision 1.18  2006/02/24 17:08:22  as-age
// Fix missing parameter.
//
// Revision 1.17  2006/02/21 20:26:50  as-age
// Fix for updated tHPIControlCacheValue structure.
//
// Revision 1.16  2006/01/13 01:30:19  as-ewb
// tweaks to make compatible with old and new hpi api.
//
// Revision 1.15  2005/12/14 00:26:18  as-sgt
// SGT - only program if 1 adapter is found on the network (until we get IDs)
//
// Revision 1.14  2005/11/07 23:36:01  as-ewb
// ewb send special "reset" adapter close message when finished if any images have been programmed
//
// Revision 1.13  2005/08/30 21:26:12  as-ewb
// ewb exit if subsysopen fails
//
// Revision 1.12  2005/08/19 15:54:59  as-age
// AGE - add code to read changed control array from a ASI2416. Seems to work.
//
// Revision 1.11  2005/08/10 07:30:57  as-ewb
// ewb select adapter from commandline by index or mac address
//
// Revision 1.10  2005/08/05 20:15:48  as-age
// AGE - add code to dump the returned control array - seems to work !
//
// Revision 1.9  2005/08/05 01:10:57  as-ewb
// ewb use a table for firmware details.
//
// Revision 1.8  2005/08/04 13:45:55  as-age
// AGE - tweaks for modifications to hpinet.h structure defns for DSP compiler support.
// No functional changes.
//
// Revision 1.7  2005/08/04 03:29:13  as-ewb
// ewb update to match hpinet.c.
// Add test code for get control array function.
//
// Revision 1.6  2005/08/03 22:12:39  as-ewb
// EWB tweaks for compiler warnings
//
// Revision 1.5  2005/08/03 08:27:04  as-ewb
// EWB uncomment more code needed for firmware update.
// Uncomment mixer dump call.
//
// Revision 1.4  2005/08/03 08:22:37  as-ewb
// ewb move some stuff to hpinet.c.  add some code to test hpi messages.
//
// Revision 1.3  2005/08/01 09:29:24  as-ewb
// EWB add code to get local MAC address.   Required use of pcap-int.h to
// access "opaque" pcap_t data structure.  Would like a cleaner way to do this.
// Changed default action to do nothing other than display adapter info.
//
// Revision 1.2  2005/07/28 05:12:25  as-ewb
// ewb add commandline parameter to specify which image to download.
// Default or 1 is for update image, 2=factory, 4=bootloader.  Add them to download more than one.
//
// Revision 1.1  2005/07/27 05:13:23  as-ewb
// EWB prototype firmware downloader  for ASI2400.
// Requres winpcap library.
//

